import logging

from wbtools.db.expression import WBExpressionDBManager
from wbtools.db.generic import WBGenericDBManager
from wbtools.db.person import WBPersonDBManager

logger = logging.getLogger(__name__)


class WBDBManager(object):

    def __init__(self, dbname, user, password, host):
        self.generic = WBGenericDBManager(dbname, user, password, host)
        self.expression = WBExpressionDBManager(dbname, user, password, host)
        self.paper = WBExpressionDBManager(dbname, user, password, host)
        self.person = WBPersonDBManager(dbname, user, password, host)



import logging
import psycopg2

from wbtools.db.abstract_manager import AbstractWBDBManager

logger = logging.getLogger(__name__)


class WBGenericDBManager(AbstractWBDBManager):

    def __init__(self, dbname, user, password, host):
        super().__init__(dbname, user, password, host)

    def get_all_paper_ids(self, added_or_modified_after: str = '1970-010-1'):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT DISTINCT joinkey, pap_timestamp from pap_electronic_path WHERE pap_timestamp > %s "
                         "ORDER BY pap_timestamp DESC",
                         (added_or_modified_after, ))
            res = curs.fetchall()
            return [row[0] for row in res] if res else []

    def get_curated_variations(self, exclude_id_used_as_name: bool = False):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT * FROM obo_data_variation WHERE obo_data_variation ~ 'name' AND "
                         "obo_data_variation ~ 'gene'")
            res = curs.fetchall()
            variations = []
            for row in res:
                for line in row[1].split('\n'):
                    if line.startswith('name:'):
                        var_name = line[7:-1]
                        if not exclude_id_used_as_name or not var_name.startswith("WBVar"):
                            variations.append(var_name)
                        break
                else:
                    logger.warning(f"Possible bogus variation entry in DB: {row[0]}")
            return variations



import os
import re
import ssl
import urllib.request


def skip_ssl_validation():
    if not os.environ.get('PYTHONHTTPSVERIFY', '') and getattr(ssl, '_create_unverified_context', None):
        ssl._create_default_https_context = ssl._create_unverified_context


def get_cgc_allele_designations():
    skip_ssl_validation()
    res = urllib.request.urlopen('https://cgc.umn.edu/laboratories')
    html_content = res.read().decode('utf-8')
    results = re.findall(r'<td><a href="https://cgc\.umn\.edu/laboratory/[A-Z]{1,3}">[A-Z]{1,3}</a></td>\n[ ]*<td>([a-z]+)</td>', html_content)
    return results



import itertools
import re
from dataclasses import dataclass
from enum import Enum
from typing import List

import regex
from gensim.models import KeyedVectors
from nltk import word_tokenize
from nltk.corpus import stopwords
from gensim.corpora import Dictionary
from gensim.models import WordEmbeddingSimilarityIndex
from gensim.similarities import SoftCosineSimilarity, SparseTermSimilarityMatrix
from nltk import sent_tokenize, word_tokenize
from nltk.corpus import stopwords

from wbtools.lib.scraping import get_cgc_allele_designations


stop_words = set(stopwords.words('english'))

ALL_VAR_REGEX = r'(' + '|'.join(get_cgc_allele_designations()) + '|m|p|It)(_)?([A-z]+)?([0-9]+)([a-zA-Z]{1,4}[0-9]*)?' \
                                                                 '(\[[0-9]+\])?([a-zA-Z]{1,4}[0-9]*)?(\[.+\])?'

NEW_VAR_REGEX = r'[\(\s](' + '|'.join(get_cgc_allele_designations()) + r'|m|p)([0-9]+)((' + \
                '|'.join(get_cgc_allele_designations()) + r'|m|p|ts|gf|lf|d|sd|am|cs)[0-9]+)?[\)\s\[]'


VAR_EXCLUSION_LIST = ["p53", "p35", "p21", "p38", "p68", "p120", "p130", "p107", "uv1", "w1", "u1"]


VAR_FALSE_POSITIVE_REGEX = r'(^(and|www|ab)\d+)|(al\d+$)|(^yk\d+)'


VAR_FROM_UNWANTED_PRJ_REGEX = r'^(ok|gk|tm|cx|tt)\d+$'


@dataclass
class SimilarityResult:
    score: int = 0
    paper_id: str = ''
    match_idx: int = 0
    query_idx: int = 0
    match: str = ''
    query: str = ''


class PaperSections(str, Enum):
    INTRODUCTION = 1
    METHOD = 2
    RELATED_WORK = 3
    RESULTS = 4
    DISCUSSION = 5
    CONCLUSION = 6
    ACKNOWLEDGEMENTS = 7
    REFERENCES = 8


PAPER_SECTIONS = {
    PaperSections.INTRODUCTION: [["introduction"], ["the", "section"], ["section"], 0.2],
    PaperSections.METHOD: [["method", "methods"], ["section"], ["section"], 0.3],
    PaperSections.RELATED_WORK: [["related work", "related works"], ["the", "previous", "section"], ["section"], 0.3],
    PaperSections.RESULTS: [["results"], ["section"], ["section"], 0.5],
    PaperSections.DISCUSSION: [["discussion", "discussions"], ["section"], ["section"], 0.7],
    PaperSections.CONCLUSION: [["conclusion", "conclusions"], ["the", "section"], ["section"], 0.8],
    PaperSections.ACKNOWLEDGEMENTS: [["acknowledgements"], ["the", "section"], ["section"], 0.8],
    PaperSections.REFERENCES: [["references", "literature"], ["the"], [], 0.9]
}


def get_documents_from_text(text: str, split_sentences: bool = False, remove_sections: List[PaperSections] = None,
                            must_be_present: List[PaperSections] = None) -> \
        list:
    if remove_sections:
        text = remove_sections_from_text(text=text, sections_to_remove=remove_sections, must_be_present=must_be_present)
    return sent_tokenize(text) if split_sentences else [text]


def remove_sections_from_text(text: str, sections_to_remove: List[PaperSections] = None,
                              must_be_present: List[PaperSections] = None):
    if sections_to_remove:
        sections_idx = {}
        for section, (section_matches, prefix_to_exclude, postfix_to_exclude, expected_position) in \
                PAPER_SECTIONS.items():
            sections_idx[section] = -1
            pre_match_regex = r"(?<!" + r"\s*|".join(prefix_to_exclude) + r"\s*)" if prefix_to_exclude else ""
            post_match_regex = r"(?!\s*" + r"\s*|".join(postfix_to_exclude) + ")" if postfix_to_exclude else ""
            for string_to_match in section_matches:
                section_match_regex = pre_match_regex + "(" + string_to_match.title() + "|" + \
                                      string_to_match.upper() + ")" + post_match_regex
                matches_idx = [match.start() for match in regex.finditer(section_match_regex, text)]
                if matches_idx:
                    # take the index closest to the expected position, if it's les than 50% off the expected position
                    idx_sorted = sorted([(match_id, match_id / len(text) - expected_position) for match_id in
                                         matches_idx if match_id / len(text) - expected_position < 0.5],
                                        key=lambda x: abs(x[1]))
                    if idx_sorted:
                        sections_idx[section] = idx_sorted[0][0]
                        if sections_idx[section] != -1:
                            break
        if not must_be_present or all([sections_idx[must_sec] != -1 for must_sec in must_be_present]):
            sec_idx_arr = sorted([(section, idx) for section, idx in sections_idx.items() if idx > -1],
                                 key=lambda x: x[1])
            # remove sections too close to each other (e.g., RESULTS AND DISCUSSION)
            sec_idx_arr = [s_i for i, s_i in enumerate(sec_idx_arr) if i == 0 or s_i[1] - sec_idx_arr[i - 1][1] > 20]
            # get first and last position of each section based on the next one
            slices_idx = [(s_i[1], (sec_idx_arr[i + 1][1] - 1) if i < len(sec_idx_arr) - 1 else len(text)) for
                          i, s_i in enumerate(sec_idx_arr) if s_i[0] in sections_to_remove]
            text = "".join([c for i, c in enumerate(text) if not any([slice_idx[0] <= i <= slice_idx[1] for slice_idx in
                                                                      slices_idx])])
    return text


def preprocess(doc, lower: bool = False, tokenize: bool = False, remove_stopwords: bool = False,
               remove_alpha: bool = False):
    if lower:
        doc = doc.lower()
    if tokenize:
        doc = word_tokenize(doc)
    if remove_stopwords:
        doc = [w for w in doc if w not in stop_words]  # Remove stopwords.
    if remove_alpha:
        doc = [w for w in doc if w.isalpha()]  # Remove numbers and special characters
    return doc


def get_softcosine_index(model_path: str = '', model=None, corpus_list_token: list = None, num_best: int = 10):
    if model_path:
        model = KeyedVectors.load_word2vec_format(model_path, binary=True)
    elif model:
        model = model.wv
    else:
        raise Exception('no model or model path provided')
    termsim_index = WordEmbeddingSimilarityIndex(model)
    dictionary = Dictionary(corpus_list_token)
    bow_corpus = [dictionary.doc2bow(doc) for doc in corpus_list_token]
    similarity_matrix = SparseTermSimilarityMatrix(termsim_index, dictionary)
    return SoftCosineSimilarity(bow_corpus, similarity_matrix, num_best=num_best), dictionary


def get_similar_documents(similarity_index, dictionary, query_documents, idx_filename_map, average_match: bool = True):
    if average_match:
        res = [SimilarityResult(score=sim[1], paper_id=idx_filename_map[sim[0]], match_idx=sim[0], query_idx=-1)
               for sim in similarity_index[dictionary.doc2bow(list(itertools.chain.from_iterable(query_documents)))]]
    else:
        res = [SimilarityResult(score=sim[1], paper_id=idx_filename_map[sim[0]], match_idx=sim[0], query_idx=q_idx) for
               q_idx, query_doc in enumerate(query_documents) for sim in similarity_index[dictionary.doc2bow(
                query_doc)]]
    return sorted(res, key=lambda x: x.score, reverse=True)


def get_entities_from_text(text, regex):
    res = re.findall(regex, " " + text + " ")
    return ["".join(entity_arr) for entity_arr in res]


def is_variation_suspicious(variation):
    return "suspicious" if re.match(r"^p\d+$", variation) or re.match(r"^(c|m|n|k)?m[1-3]$", variation) or \
            re.match(r"^ws\d+$", variation) or re.match(r"^aa\d+", variation) or \
            re.match(r"^x(1|2|3|100|200|500|1000)$", variation) else "potentially new"


def get_new_variations_from_text(text):
    variations = set(get_entities_from_text(text, NEW_VAR_REGEX))
    variations = [var for var in variations if var not in VAR_EXCLUSION_LIST and not re.match(
        VAR_FALSE_POSITIVE_REGEX, var) and not re.match(VAR_FROM_UNWANTED_PRJ_REGEX, var)]
    return [(variation, is_variation_suspicious(variation)) for variation in variations]





import psycopg2

from wbtools.db.abstract_manager import AbstractWBDBManager
from wbtools.literature.person import WBPerson


class WBPersonDBManager(AbstractWBDBManager):

    def __init__(self, dbname, user, password, host):
        super().__init__(dbname, user, password, host)

    def get_person_id_from_email_address(self, email_address):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT * FROM two_email WHERE two_email=%s", (email_address, ))
            res = curs.fetchone()
            if res:
                return res[0]
            else:
                curs.execute("SELECT * FROM two_old_email WHERE two_old_email=%s", (email_address, ))
                res = curs.fetchone()
                if res:
                    return res[0]
            return None

    def get_first_name(self, person_id):
        return self._get_single_field(person_id, "two_firstname")

    def get_middle_name(self, person_id):
        return self._get_single_field(person_id, "two_middlename")

    def get_last_name(self, person_id):
        return self._get_single_field(person_id, "two_lastname")

    def get_aka_first_name(self, person_id):
        return self._get_single_field(person_id, "two_aka_firstname")

    def get_aka_middle_name(self, person_id):
        return self._get_single_field(person_id, "two_aka_middlename")

    def get_aka_last_name(self, person_id):
        return self._get_single_field(person_id, "two_aka_lastname")

    def get_orcid(self, person_id):
        return self._get_single_field(person_id, "two_orcid")

    def get_city(self, person_id):
        return self._get_single_field(person_id, "two_city")

    def get_state(self, person_id):
        return self._get_single_field(person_id, "two_state")

    def get_street(self, person_id):
        return self._get_single_field(person_id, "two_street")

    def get_country(self, person_id):
        return self._get_single_field(person_id, "two_country")

    def get_email(self, person_id):
        return self._get_single_field(person_id, "two_email")

    def get_fax(self, person_id):
        return self._get_single_field(person_id, "two_fax")

    def get_institution(self, person_id):
        return self._get_single_field(person_id, "two_institution")

    def get_lab(self, person_id):
        return self._get_single_field(person_id, "two_lab")

    def get_old_lab(self, person_id):
        return self._get_single_field(person_id, "two_oldlab")

    def get_lab_phone(self, person_id):
        return self._get_single_field(person_id, "two_labphone")

    def get_status(self, person_id):
        return self._get_single_field(person_id, "two_status")

    def get_main_phone(self, person_id):
        return self._get_single_field(person_id, "two_mainphone")

    def get_office_phone(self, person_id):
        return self._get_single_field(person_id, "two_officephone")

    def get_old_emails(self, person_id):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT two_old_email FROM two_old_email WHERE joinkey=%s", (person_id, ))
            res = curs.fetchall()
            return [row[0] for row in res]

    def get_comment(self, person_id):
        return self._get_single_field(person_id, "two_comment")

    def get_contact_data(self, person_id):
        return self._get_single_field(person_id, "two_contactdata")

    def get_person(self, person_id) -> WBPerson:
        person = WBPerson()
        person.person_id = person_id
        person.first_name = self.get_first_name(person_id)
        person.middle_name = self.get_middle_name(person_id)
        person.last_name = self.get_last_name(person_id)
        person.aka_first_name = self.get_aka_first_name(person_id)
        person.aka_middle_name = self.get_aka_middle_name(person_id)
        person.aka_last_name = self.get_aka_last_name(person_id)
        person.orcid = self.get_orcid(person_id)
        person.city = self.get_city(person_id)
        person.state = self.get_state(person_id)
        person.street = self.get_street(person_id)
        person.country = self.get_country(person_id)
        person.email = self.get_email(person_id)
        person.fax = self.get_fax(person_id)
        person.institution = self.get_institution(person_id)
        person.lab = self.get_lab(person_id)
        person.old_lab = self.get_old_lab(person_id)
        person.lab_phone = self.get_lab_phone(person_id)
        person.status = self.get_status(person_id)
        person.main_phone = self.get_main_phone(person_id)
        person.office_phone = self.get_office_phone(person_id)
        person.old_emails = self.get_old_emails(person_id)
        person.comment = self.get_comment(person_id)
        person.contact_data = self.get_contact_data(person_id)
        return person

    def get_author_corresponding(self, author_id):
        return self._get_single_field(author_id, "pap_author_corresponding") == "corresponding"



import abc
from typing import Union

import psycopg2


class AbstractWBDBManager(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self, dbname, user, password, host):
        self.connection_str = "dbname='" + dbname
        if user:
            self.connection_str += "' user='" + user
        if password:
            self.connection_str += "' password='" + password
        self.connection_str += "' host='" + host + "'"

    def _get_single_field(self, join_key: str, field_name: str) -> Union[str, None]:
        """
        get a specific field from a single field table. Field and table name must be equal

        Args:
            join_key (str): joinkey
            field_name (str): the field to retrieve from the related <field_name> table
        Returns:
            str: the value of the specified field
        """
        table_prefix = field_name.split("_")[0]
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT {} FROM {} WHERE joinkey = %s ORDER BY {}_timestamp DESC".format(
                field_name, field_name, table_prefix), (join_key,))
            res = curs.fetchone()
            return res[0] if res and res[0] != "NULL" and res[0] != "null" else None



from typing import List


class WBPerson(object):

    def __init__(self, person_id: str = None, first_name: str = None, last_name: str = None, middle_name: str = None,
                 aka_first_name: str = None, aka_middle_name: str = None, aka_last_name: str = None, city: str = None,
                 state: str = None, orcid: str = None, street: str = None,
                 country: str = None, email: str = None, fax: str = None,
                 institution: str = None, lab: str = None, old_lab: str = None, lab_phone: str = None,
                 status: bool = True, main_phone: str = None, office_phone: str = None, old_emails: List[str] = None,
                 comment: str = None, contact_data: str = None):
        self.person_id = person_id
        self.first_name = first_name
        self.middle_name = middle_name
        self.last_name = last_name
        self.aka_first_name = aka_first_name
        self.aka_middle_name = aka_middle_name
        self.aka_last_name = aka_last_name
        self.orcid = orcid
        self.city = city
        self.state = state
        self.street = street
        self.country = country
        self.email = email
        self.fax = fax
        self.institution = institution
        self.lab = lab
        self.old_lab = old_lab
        self.lab_phone = lab_phone
        self.status = status
        self.main_phone = main_phone
        self.office_phone = office_phone
        self.old_emails = old_emails
        self.comment = comment
        self.contact_data = contact_data


class WBAuthor(WBPerson):

    def __init__(self, *args, corresponding: bool = False, verified: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.corresponding = corresponding
        self.verified = verified

    @classmethod
    def from_person(cls, person):
        return cls(**{k: v for k, v in person.__dict__.items() if not k.startswith('__') and not callable(k)})



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wbtools",
    version="1.0.2",
    author="Valerio Arnaboldi",
    author_email="valearna@caltech.edu",
    description="Interface to WormBase (www.wormbase.org) curation data, including literature management and NLP "
                "functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/WormBase/wbtools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'psycopg2-binary',
        'numpy~=1.19.2',
        'fabric~=2.5.0',
        'gensim~=3.8.3',
        'nltk~=3.5',
        'setuptools~=50.3.2',
        'PyMuPDF~=1.18.4'
    ]
)



from typing import Union, List
from psycopg2 import sql

import psycopg2

from wbtools.db.abstract_manager import AbstractWBDBManager
from wbtools.db.person import WBPersonDBManager
from wbtools.literature.person import WBAuthor


class WBPaperDBManager(AbstractWBDBManager):

    def __init__(self, dbname, user, password, host):
        super().__init__(dbname, user, password, host)
        self.person_db_manager = WBPersonDBManager(dbname, user, password, host)

    def get_paper_abstract(self, paper_id):
        return self._get_single_field(join_key=paper_id, field_name="pap_abstract")

    def get_paper_title(self, paper_id):
        return self._get_single_field(join_key=paper_id, field_name="pap_title")

    def get_paper_journal(self, paper_id):
        return self._get_single_field(join_key=paper_id, field_name="pap_journal")

    def get_paper_pub_date(self, paper_id):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute(
                "SELECT pap_year, pap_month, pap_day FROM pap_year JOIN pap_month ON "
                "pap_year.joinkey = pap_month.joinkey JOIN pap_day ON pap_year.joinkey = pap_day.joinkey "
                "WHERE pap_year.joinkey = %s", (paper_id,))
            res = curs.fetchone()
            return res[0] + '-' + res[1] + '-' + res[2] if res else None

    def get_paper_authors(self, paper_id) -> List[WBAuthor]:
        authors = []
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute(
                "SELECT pap_author FROM pap_author "
                "WHERE joinkey = %s", (paper_id,))
            res = curs.fetchall()
            authors_ids = [row[0] for row in res]
            for author_id in authors_ids:
                curs.execute("SELECT pap_author_possible, pap_author_verified LIKE 'YES%%' "
                             "FROM pap_author_possible LEFT OUTER JOIN pap_author_verified ON "
                             "pap_author_possible.author_id = pap_author_verified.author_id and "
                             "pap_author_possible.pap_join = pap_author_verified.pap_join "
                             "WHERE pap_author_possible.author_id = %s",
                             (author_id,))
                res = curs.fetchone()
                if res:
                    person = self.person_db_manager.get_person(person_id=res[0])
                    author = WBAuthor.from_person(person)
                    author.verified = res[1]
                    authors.append(author)
        return authors

    def get_pmid(self, paper_id):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute(
                "SELECT * FROM pap_identifier WHERE joinkey = %s AND pap_identifier LIKE 'pmid%'", (paper_id, ))
            res = curs.fetchone()
            if res and res[1].startswith("pmid"):
                return res[1].replace("pmid", "")
            else:
                return None

    def get_corresponding_emails(self, paper_id):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT two_email.two_email "
                         "FROM pap_author_corresponding "
                         "JOIN pap_author_possible "
                         "ON pap_author_corresponding.author_id = pap_author_possible.author_id "
                         "JOIN pap_author on pap_author.pap_author = pap_author_corresponding.author_id "
                         "JOIN two_email ON two_email.joinkey = pap_author_possible.pap_author_possible "
                         "WHERE pap_author.joinkey = %s", (paper_id,))
            res = curs.fetchone()
            return [res[0]] if res else []

    def get_svm_values(self, paper_id: str):
        """
        get all the SVM (Support Vector Machine) classifier values for the paper

        Args:
            paper_id (str): 8 digit numeric id that identifies the paper
        Returns:
            Union[str, None]: the value associated to the svm ('low', 'medium', or 'high') or None if the paper has not
                              been classified by SVMs yet
        """
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT cur_datatype, cur_svmdata from cur_svmdata WHERE cur_paper = %s", (paper_id,))
            res = curs.fetchall()
            return [(row[0], row[1]) for row in res]

    def get_file_paths(self, paper_id: str):
        """
        get the list of file paths associated with the specified paper_id

        Args:
            paper_id (str): 8 digit numeric id that identifies the paper
        Returns:
            list[str]: a list of electronic paths to the files
        """
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT pap_electronic_path FROM pap_electronic_path WHERE joinkey = %s ORDER BY joinkey",
                         (paper_id, ))
            rows = curs.fetchall()
            return [row[0] for row in rows] if rows else []



import psycopg2

from wbtools.db.abstract_manager import AbstractWBDBManager


class WBExpressionDBManager(AbstractWBDBManager):

    def __init__(self, dbname, user, password, host):
        super().__init__(dbname, user, password, host)

    def get_all_expression_sentences(self, min_num_words: int = 0):
        with psycopg2.connect(self.connection_str) as conn, conn.cursor() as curs:
            curs.execute("SELECT exp_pattern FROM exp_pattern")
            res = curs.fetchall()
            return [row[0] for row in res if len(row[0].split()) > min_num_words]



import logging
import os
import pickle

from typing import Generator, List, Tuple, Dict
from gensim.models import Word2Vec
from gensim.test.utils import common_texts

from wbtools.db.dbmanager import WBDBManager
from wbtools.lib.nlp import PaperSections, get_softcosine_index, preprocess, get_similar_documents, SimilarityResult
from wbtools.literature.paper import WBPaper


logger = logging.getLogger(__name__)


class CorpusManager(object):
    """manage a list of WBPaper objects by populating their data from database or local directory"""

    def __init__(self):
        self.corpus = {}

    def add_or_update_wb_paper(self, wb_paper: WBPaper):
        """add a paper

        Args:
            wb_paper(WBPaper): the paper to add to the corpus
        """
        self.corpus[wb_paper.paper_id] = wb_paper

    def remove_wb_paper(self, wb_paper: WBPaper):
        """remove a paper

        Args:
            wb_paper(WBPaper): the paper to remove from the corpus
        """
        del self.corpus[wb_paper.paper_id]

    def load_from_dir_with_txt_files(self, dir_path: str):
        """
        load papers from a directory containing text files with file name in the following format:
        <WBPaperID>_<Author><Year>_<additional_options>.txt

        Only files with .txt extension are loaded. Paper ID is derived from the file name and additional options are
        used to understand the type of file (e.g., main article, ocr scanned article, supplementary material etc.)

        Args:
            dir_path (str): path to the input directory containing text files
        """
        paper = WBPaper()
        for f in sorted(os.listdir(dir_path)):
            if os.path.isfile(os.path.join(dir_path, f)) and f.endswith(".txt"):
                if paper.paper_id and not paper.has_same_wbpaper_id_as_filename(f):
                    self.add_or_update_wb_paper(paper)
                    paper = WBPaper()
                paper.add_file(dir_path=dir_path, filename=f, remote_file=False, pdf=False)

    def load_from_wb_database(self, db_name: str, db_user: str, db_password: str, db_host: str,
                              tazendra_ssh_user: str = None, tazendra_ssh_passwd: str = None, paper_ids: list = None,
                              from_date: str = None, load_pdf_files: bool = True, load_bib_info: bool = True,
                              load_curation_info: bool = True, max_num_papers: int = None,
                              must_have_svm: bool = False) -> None:
        """load papers from WormBase database

        Args:
            db_name (str): database name
            db_user (str): database user
            db_password (str): database password
            db_host (str): database host
            tazendra_ssh_user (str): ssh user to fetch pdf files
            tazendra_ssh_passwd (str): ssh password to fetch pdf files
            paper_ids (list): optional list of paper ids to be fetched
            from_date (str): load papers added or modified from the specified date (only if paper_ids is not provided)
            load_pdf_files (bool): load pdf files using ssh credentials
            load_bib_info (bool): load bibliographic info of the papers
            load_curation_info (bool): load curation info of the papers
            max_num_papers (int): limit number of papers to be loaded
            must_have_svm (bool): whether to load only papers that have been flagged by WB SVMs
        """
        if not paper_ids:
            db_manager = WBDBManager(db_name, db_user, db_password, db_host)
            paper_ids = db_manager.generic.get_all_paper_ids(added_or_modified_after=from_date)
        for paper_id in paper_ids:
            paper = WBPaper(paper_id=paper_id, tazendra_ssh_user=tazendra_ssh_user,
                            tazendra_ssh_passwd=tazendra_ssh_passwd)
            if load_curation_info:
                paper.load_curation_info_from_db(db_name=db_name, db_user=db_user, db_password=db_password,
                                                 db_host=db_host)
                if must_have_svm and not paper.svm_values:
                    continue
            if load_pdf_files:
                paper.load_text_from_pdf_files_in_db(db_name=db_name, db_user=db_user, db_password=db_password,
                                                     db_host=db_host)
            if load_bib_info:
                paper.load_bib_info_from_db(db_name=db_name, db_user=db_user, db_password=db_password, db_host=db_host)
            self.add_or_update_wb_paper(paper)
            if max_num_papers and self.size() >= max_num_papers:
                break

    def size(self) -> int:
        """number of papers in the corpus manager

        Returns:
            int: the number of papers
        """
        return len(self.corpus)

    def get_flat_corpus_list_and_idx_paperid_map(self, split_sentences: bool = False,
                                                 remove_sections: List[PaperSections] = None,
                                                 must_be_present: List[PaperSections] = None,
                                                 lowercase: bool = False, tokenize: bool = False,
                                                 remove_stopwords: bool = False,
                                                 remove_alpha: bool = False) -> Tuple[List[str], Dict[int, str]]:
        """get a flat list of text documents from the papers in the corpus and a map to link the index in the resulting
           list and the id of the related paper

        Args:
            split_sentences (bool): split sentences into separate documents
            remove_sections (List[PaperSections]): list of sections to remove
            must_be_present (List[PaperSections]): list of sections that must be present
            lowercase (bool): transform text to lowercase
            tokenize (bool): tokenize text into words
            remove_stopwords (bool): remove common stopwords from text
            remove_alpha (bool): remove special characters and punctuation from text

        Returns:
            Tuple[List[str], Dict[int, str]]: the flat list and the related index to paper id map
        """
        flat_list_with_ids = [(doc, paper.paper_id) for paper in self.corpus.values() for doc in paper.get_text_docs(
            include_supplemental=True, remove_sections=remove_sections, must_be_present=must_be_present,
            split_sentences=split_sentences, lowercase=lowercase, tokenize=tokenize, remove_stopwords=remove_stopwords,
            remove_alpha=remove_alpha)]
        return [d[0] for d in flat_list_with_ids], {idx: d[1] for idx, d in enumerate(flat_list_with_ids)}

    def get_paper(self, paper_id) -> WBPaper:
        """get a paper from the corpus by paper id

        Args:
            paper_id (str): paper id to retrieve

        Returns:
            WBPaper: the paper
        """
        return self.corpus[paper_id]

    def get_all_papers(self) -> Generator[WBPaper, None, None]:
        """get all the papers in the corpus

        Returns:
            Generator[WBPaper, None, None]: a generator to the papers in the corpus
        """
        for paper in self.corpus.values():
            yield paper

    def save(self, file_path: str) -> None:
        """save corpus to file

        Args:
            file_path (str): path to file to save
        """
        with open(file_path, 'wb') as out_file:
            pickle.dump(self, out_file)

    def load(self, file_path: str) -> None:
        """load corpus from previously saved file

        Args:
            file_path (str): path to file to load
        """
        with open(file_path, 'rb') as in_file:
            tmp_self = pickle.load(in_file)
            self.__dict__ = tmp_self.__dict__

    def query_papers_by_doc_similarity(self, query_docs: List[str], sentence_search: bool = False,
                                       remove_sections: List[PaperSections] = None,
                                       must_be_present: List[PaperSections] = None, path_to_model: str = None,
                                       average_match: bool = True, num_best: int = 10) -> List[SimilarityResult]:
        """query papers in the corpus by similarity with the provided query documents, which can be fulltext documents
           or sentences

        Args:
            query_docs (List[str]): list of query documents
            sentence_search (bool): perform sentence level similarity search
            remove_sections (List[PaperSections]): sections to be ignored from corpus papers
            must_be_present (List[PaperSections]): sections that must be present in corpus papers before removing
                                                   sections
            path_to_model (str): path to word2vec model
            average_match (bool): merge query documents and calculate average similarity to them
            num_best (int): limit to the first n results by similarity score

        Returns:
            List[SimilarityResult]: list of papers most similar to the provided query documents
        """
        model = Word2Vec(common_texts, min_count=1) if not path_to_model else None
        corpus_list_token, idx_paperid_map = self.get_flat_corpus_list_and_idx_paperid_map(
            split_sentences=sentence_search, remove_sections=remove_sections, must_be_present=must_be_present,
            lowercase=True, tokenize=True, remove_stopwords=True, remove_alpha=True)
        corpus_list_token_orig, _ = self.get_flat_corpus_list_and_idx_paperid_map(
            split_sentences=sentence_search, remove_sections=remove_sections, must_be_present=must_be_present,
            lowercase=False, tokenize=False, remove_stopwords=False, remove_alpha=False)
        docsim_index, dictionary = get_softcosine_index(model=model, model_path=path_to_model,
                                                        corpus_list_token=corpus_list_token, num_best=num_best)
        query_docs_preprocessed = [preprocess(doc=sentence, lower=True, tokenize=True, remove_stopwords=True,
                                              remove_alpha=True) for sentence in query_docs]
        sims = get_similar_documents(docsim_index, dictionary, query_docs_preprocessed, idx_paperid_map,
                                     average_match=average_match)
        results = [SimilarityResult(score=sim.score, paper_id=sim.paper_id, match_idx=sim.match_idx,
                                    query_idx=sim.query_idx, match="\"" + corpus_list_token_orig[sim.match_idx] + "\"",
                                    query="\"" + (" ".join(query_docs) if average_match else query_docs[sim.query_idx]
                                                  ) + "\"") for sim in sims]
        return results[0:num_best] if len(results) > num_best else results



import logging
import os
import re
import tempfile
from typing import List

import numpy as np

from collections import defaultdict
import fitz
from fabric.connection import Connection

from wbtools.db.paper import WBPaperDBManager
from wbtools.lib.nlp import preprocess, get_documents_from_text, PaperSections
from wbtools.lib.timeout import timeout
from wbtools.literature.person import WBAuthor

logger = logging.getLogger(__name__)


TAZENDRA_SSH_HOST = "tazendra.caltech.edu"


class PaperFileReader(object):

    def __init__(self, tazendra_ssh_user: str = '', tazendra_ssh_passwd: str = ''):
        self.tazendra_ssh_user = tazendra_ssh_user
        self.tazendra_ssh_passwd = tazendra_ssh_passwd

    @staticmethod
    @timeout(3600)
    def convert_pdf_to_txt(file_path):
        logger.info("Started pdf to text conversion")
        doc = fitz.Document(file_path)
        text = " ".join([doc.loadPage(page_num).getText("text") for page_num in range(doc.pageCount)])
        return text if text is not None else ""

    def get_supplemental_file_names(self, supp_dir_path):
        with Connection(TAZENDRA_SSH_HOST, self.tazendra_ssh_user,
                        connect_kwargs={"password": self.tazendra_ssh_passwd}) as c, c.sftp() as sftp:
            try:
                return [filename for filename in sftp.listdir(supp_dir_path) if filename.endswith(".pdf")]
            except UnicodeDecodeError:
                logger.error("Cannot read non-unicode chars in filenames due to bug in ssh library")
                return []

    def download_paper_and_extract_txt(self, file_url, pdf: bool = False):
        with Connection(TAZENDRA_SSH_HOST, self.tazendra_ssh_user, connect_kwargs={"password": self.tazendra_ssh_passwd}) as \
                c, c.sftp() as sftp, sftp.open(file_url) as file_stream:
            tmp_file = tempfile.NamedTemporaryFile()
            with open(tmp_file.name, 'wb') as tmp_file_stream:
                tmp_file_stream.write(file_stream.read())
        if pdf:
            return self.convert_pdf_to_txt(tmp_file.name)
        else:
            return open(tmp_file.name).read()

    def get_text_from_file(self, dir_path, filename, remote_file: bool = False, pdf: bool = False):
        if remote_file:
            text = self.download_paper_and_extract_txt(dir_path + filename, pdf)
        else:
            with open(os.path.join(dir_path, filename), 'r') as file:
                text = file.read()
        if np.average([len(w) for w in preprocess(text).split("\n")]) < 1.001:
            text = text.replace("\n\n", " ")
            text = text.replace("\n", "")
        else:
            text = text.replace("\n", " ")
        text = text.replace("Fig.", "Fig")
        text = text.replace("et al.", "et al")
        return text


class WBPaper(object):
    """WormBase paper information"""

    def __init__(self, paper_id: str = '', main_text: str = '', ocr_text: str = '', temp_text: str = '',
                 aut_text: str = '', html_text: str = '', supplemental_docs: list = None, tazendra_ssh_user: str = '',
                 tazendra_ssh_passwd: str = '', title: str = '', journal: str = '', pub_date: str = '',
                 authors: List[WBAuthor] = None, abstract: str = ''):
        self.paper_id = paper_id
        self.title = title
        self.journal = journal
        self.pub_date = pub_date
        self.authors = authors
        self.abstract = abstract
        self.main_text = main_text
        self.ocr_text = ocr_text
        self.temp_text = temp_text
        self.aut_text = aut_text
        self.html_text = html_text
        self.supplemental_docs = supplemental_docs if supplemental_docs else []
        self.svm_values = defaultdict(str)
        self.paper_file_reader = PaperFileReader(tazendra_ssh_user=tazendra_ssh_user,
                                                 tazendra_ssh_passwd=tazendra_ssh_passwd)

    def get_corresponding_author(self):
        for author in self.authors:
            if author.corresponding:
                return author
        return None

    def get_text_docs(self, include_supplemental: bool = True, remove_sections: List[PaperSections] = None,
                      must_be_present: List[PaperSections] = None, split_sentences: bool = False,
                      lowercase: bool = False, tokenize: bool = False, remove_stopwords: bool = False,
                      remove_alpha: bool = False) -> list:
        """get text documents for the paper

        Args:
            include_supplemental (bool): include supplemental material
            remove_sections (List[PaperSections]): sections to be removed
            must_be_present (List[PaperSections]): sections that must be present in the text to be able to remove
                                                   undesired sections
            split_sentences (bool): split documents into sections
            lowercase (bool): transform text to lowercase
            tokenize (bool): tokenize text into words
            remove_stopwords (bool): remove common stopwords
            remove_alpha (bool): remove special characters and punctuation

        Returns:
            list: a list of documents, which can be strings or lists of strings or list of lists of strings according to
                  the combination of tokenization arguments passed
        """
        docs = [self.main_text if self.main_text else self.html_text if self.html_text else self.ocr_text if
                self.ocr_text else self.aut_text if self.aut_text else self.temp_text]
        if include_supplemental:
            docs.extend(self.supplemental_docs)
        docs = [d for doc in docs for d in get_documents_from_text(
            text=doc, split_sentences=split_sentences, must_be_present=must_be_present,
            remove_sections=remove_sections)]
        return [preprocess(doc, lower=lowercase, tokenize=tokenize, remove_stopwords=remove_stopwords,
                           remove_alpha=remove_alpha) for doc in docs]

    def add_file(self, dir_path, filename, remote_file: bool = False, pdf: bool = False):
        """add one or more files to the paper. Information about the type of file is derived from file name. If the file
           path points to a directory, all supplementary files in it are loaded.

        Args:
            dir_path (str): path to the base directory
            filename (str): name of the file to load
            remote_file (bool): whether the file is on a remote location
            pdf (bool): whether the file is in pdf format
        """
        all_supp = False
        if not self.paper_file_reader and remote_file:
            raise Exception("a paper reader must be provided to access remote files")
        if dir_path.endswith("supplemental/") and re.match(r'^[0-9]+$', filename):
            filenames = self.paper_file_reader.get_supplemental_file_names(dir_path + filename)
            dir_path = dir_path + filename + "/"
            all_supp = True
        else:
            filenames = [filename]
        for filename in filenames:
            logger.info("Adding file " + dir_path + filename)
            if "_lib" not in filename and "Fig" not in filename and "Figure" not in filename:
                wb_paperid, author_year, additional_options = self._get_matches_from_filename(filename)
                if not self.paper_id:
                    self.paper_id = wb_paperid
                if all_supp or (author_year and ("_supp" in author_year or "_Supp" in author_year or "_Table" in
                                                 author_year or "_table" in author_year or "_mmc" in author_year or
                                                 "_Stable" in author_year or "_Movie" in author_year or "_movie" in
                                                 author_year or "supplementary" in author_year or "Supplementary" in
                                                 author_year or re.match(r'[_-][Ss][0-9]+', author_year))):
                    self.supplemental_docs.append(self.paper_file_reader.get_text_from_file(
                        dir_path, filename, remote_file, pdf))
                    return
                if not additional_options:
                    self.main_text = self.paper_file_reader.get_text_from_file(dir_path, filename, remote_file, pdf)
                elif "Supp" in additional_options or "supp" in additional_options or "Table" in additional_options or \
                        "table" in additional_options or "Movie" in additional_options or "movie" in additional_options:
                    self.supplemental_docs.append(self.paper_file_reader.get_text_from_file(
                        dir_path, filename, remote_file, pdf))
                elif "ocr" in additional_options:
                    self.ocr_text = self.paper_file_reader.get_text_from_file(dir_path, filename, remote_file, pdf)
                elif "temp" in additional_options:
                    self.temp_text = self.paper_file_reader.get_text_from_file(dir_path, filename, remote_file, pdf)
                elif "aut" in additional_options:
                    self.aut_text = self.paper_file_reader.get_text_from_file(dir_path, filename, remote_file, pdf)
                elif "html" in additional_options:
                    self.html_text = self.paper_file_reader.get_text_from_file(dir_path, filename, remote_file, pdf)
                else:
                    logger.warning("No rule to read filename: " + filename)

    def load_text_from_pdf_files_in_db(self, db_name, db_user, db_password, db_host):
        """load text from pdf files in the WormBase database

        Args:
            db_name (str): database name
            db_user (str): database user
            db_password (str): database password
            db_host (str): database host
        """
        wb_paper_db_manager = WBPaperDBManager(db_name, db_user, db_password, db_host)
        file_paths = wb_paper_db_manager.get_file_paths(self.paper_id)
        for file_path in file_paths:
            filename = file_path.split("/")[-1]
            dir_path = file_path.rstrip(filename)
            self.add_file(dir_path=dir_path, filename=filename, remote_file=True, pdf=True)

    def load_curation_info_from_db(self, db_name, db_user, db_password, db_host):
        """load curation data from WormBase database

        Args:
            db_name (str): database name
            db_user (str): database user
            db_password (str): database password
            db_host (str): database host
        """
        wb_paper_db_manager = WBPaperDBManager(db_name, db_user, db_password, db_host)
        svm_values = wb_paper_db_manager.get_svm_values(paper_id=self.paper_id)
        for svm_type, svm_value in svm_values:
            self.svm_values[svm_type] = svm_value

    def load_bib_info_from_db(self, db_name, db_user, db_password, db_host):
        """load curation data from WormBase database

        Args:
            db_name (str): database name
            db_user (str): database user
            db_password (str): database password
            db_host (str): database host
        """
        wb_paper_db_manager = WBPaperDBManager(db_name, db_user, db_password, db_host)
        self.abstract = wb_paper_db_manager.get_paper_abstract(self.paper_id)
        self.title = wb_paper_db_manager.get_paper_title(self.paper_id)
        self.journal = wb_paper_db_manager.get_paper_journal(self.paper_id)
        self.abstract = wb_paper_db_manager.get_paper_abstract(self.paper_id)
        self.pub_date = wb_paper_db_manager.get_paper_pub_date(self.paper_id)
        self.authors = wb_paper_db_manager.get_paper_authors(self.paper_id)

    def has_same_wbpaper_id_as_filename(self, filename):
        return self._get_matches_from_filename(filename)[0] == self.paper_id

    @staticmethod
    def _get_matches_from_filename(filename):
        match = re.match(r'^([0-9]+[-_])?([^_]+)?(_.*)?\..*$', filename)
        if match:
            return match.group(1)[:-1] if match.group(1) else None, match.group(2), match.group(3)
        else:
            raise Exception("Can't extract WBPaperID from filename: " + filename)

    def is_temp(self):
        """determine if the paper has a temporary pdf

        Returns:
            bool: whether the paper has a temporary pdf file
        """
        return not self.main_text

    def has_supplementary_material(self):
        """determine if the paper has supplemental material associated to it

        Returns:
            bool: whether the paper has supplemental material
        """
        return self.supplemental_docs
